#!/usr/bin/env python
# coding: utf-8

# In[17]:


#import findspark
#findspark.find()
#findspark.init()


# In[7]:


import pandas as pd
import re
from sparkkgml.data_acquisition import DataAcquisition
from kgextension.endpoints import DBpedia
from pyspark.sql import SparkSession


# In[8]:


def clean_column_names(df):
    
    """
        Clean column names of a Pandas DataFrame by removing invalid characters.

        Args:
            df (pandas.DataFrame): The input Pandas DataFrame.

        Returns:
            pandas.DataFrame: The DataFrame with cleaned column names.
    """
    
    # Regular expression pattern for removing invalid characters
    pattern = r'[\s`/\\,:;\'"\[\]\{\}\(\).]'
    
    # Apply strip to remove leading and trailing whitespace characters
    #cleaned_columns = [col.strip() for col in df]
    
    # Clean the column names
    cleaned_columns = [re.sub(pattern, '', col) for col in df]
    
    # Create a dictionary to map original column names to cleaned column names
    column_mapping = {col: cleaned_col for col, cleaned_col in zip(df.columns, cleaned_columns)}
    
    #Rename the columns in the DataFrame
    df_cleaned = df.rename(columns=column_mapping)
    
    return df_cleaned


# In[9]:


def spark_dbpedia_lookup_linker(sparkDataFrame, column, new_attribute_name='new_link', progress=True, 
                                base_url='https://lookup.dbpedia.org/api/search/', max_hits=1, query_class='', 
                                lookup_api='KeywordSearch', caching=True):
    
    """
        Perform DBpedia entity linking on a Spark DataFrame using the DBpedia Spotlight service.

        Args:
            sparkDataFrame (pyspark.sql.DataFrame): The input Spark DataFrame.
            column (str): Name of the column whose entities should be found.
            new_attribute_name (str, optional): Name of the column containing the link to the knowledge graph.
                                                Defaults to 'new_link'.
            progress (bool, optional): If True, progress bars will be shown to inform the user about the progress.
                                       Defaults to True.
            base_url (str, optional): The base URL of the DBpedia Lookup API. Defaults to 'https://lookup.dbpedia.org/api/search/'.
            max_hits (int, optional): Maximal number of URIs that should be returned per entity. Defaults to 1.
            query_class (str, optional): Specifies whether the entities that occur first ('first'), that have the highest support
                                         ('support'), or that have the highest similarity score ('similarityScore') should be chosen.
                                         Defaults to ''.
            lookup_api (str, optional): The DBpedia Lookup API to use. Defaults to 'KeywordSearch'.
            caching (bool, optional): Turn result-caching for queries issued during the execution on or off. Defaults to True.

        Returns:
            pyspark.sql.DataFrame: DataFrame with new column(s) containing the DBpedia URIs.

        Notes:
            This function performs DBpedia entity linking on a Spark DataFrame using the DBpedia Spotlight service. It follows these steps:

            1. Convert the Spark DataFrame to a Pandas DataFrame.

            2. Apply the `dbpedia_lookup_linker` function from the kgextension library to the Pandas DataFrame.

            3. Convert the resulting Pandas DataFrame back to a Spark DataFrame.

            4. Return the Spark DataFrame with the new column(s) containing the DBpedia URIs.
        """
    from kgextension.linking import dbpedia_lookup_linker
    
    spark = SparkSession.builder.getOrCreate()
    
    DataAcquisitionObject=DataAcquisition()
    
    pandasDf=dbpedia_lookup_linker(sparkDataFrame.toPandas(), column, new_attribute_name, progress, base_url, max_hits, query_class, lookup_api, caching)
    
    sparkDataFrame2= spark.createDataFrame(clean_column_names(DataAcquisitionObject.nullDrop(pandasDf)))
    
    return sparkDataFrame2


# In[10]:


def spark_specific_relation_generator(sparkDataFrame, columns, endpoint=DBpedia, uri_data_model=False, progress=True, 
                                      direct_relation="http://purl.org/dc/terms/subject", 
                                      hierarchy_relation=None, max_hierarchy_depth=1, prefix_lookup=False, caching=True):
    
    
    """
        Generate attributes from a specific direct relation on a Spark DataFrame.

        Args:
            sparkDataFrame (pyspark.sql.DataFrame): The input Spark DataFrame.
            columns (str or list): Name(s) of column(s) that contain(s) the link(s) to the knowledge graph.
            endpoint (Endpoint, optional): SPARQL Endpoint to be queried; ignored when 'uri_data_model' is True.
                                           Defaults to DBpedia.
            uri_data_model (bool, optional): If enabled, the URI is directly queried instead of a SPARQL endpoint.
                                             Defaults to False.
            progress (bool, optional): If True, progress bars will be shown to inform the user about the progress made
                                       by the process. Defaults to True.
            direct_relation (str, optional): Direct relation used to create features. Defaults to
                                             'http://purl.org/dc/terms/subject'.
            hierarchy_relation (str, optional): Hierarchy relation used to connect categories. Defaults to None.
            max_hierarchy_depth (int, optional): Maximal number of hierarchy steps taken. Defaults to 1.
            prefix_lookup (bool/str/dict, optional): True: Namespaces of prefixes will be looked up at prefix.cc and added
                                                     to the SPARQL query.
                                                     str: User provides the path to a JSON file with prefixes and namespaces.
                                                     dict: User provides a dictionary with prefixes and namespaces.
                                                     Defaults to False.
            caching (bool, optional): Turn result-caching for queries issued during the execution on or off. Defaults to True.

        Returns:
            pyspark.sql.DataFrame: DataFrame with additional features.

        Notes:
            This function generates attributes from a specific direct relation on a Spark DataFrame. It follows these steps:

            1. Convert the Spark DataFrame to a Pandas DataFrame.

            2. Apply the `specific_relation_generator` function from the kgextension library to the Pandas DataFrame.

            3. Convert the resulting Pandas DataFrame back to a Spark DataFrame.

            4. Return the Spark DataFrame with the additional features.
            
    """
    from kgextension.generator import specific_relation_generator
    
    spark = SparkSession.builder.getOrCreate()
    
    DataAcquisitionObject=DataAcquisition()

    pandasDf=specific_relation_generator(sparkDataFrame.toPandas(), columns, endpoint, uri_data_model, progress, direct_relation, hierarchy_relation, max_hierarchy_depth, prefix_lookup, caching)
    
    sparkDataFrame2= spark.createDataFrame(clean_column_names(DataAcquisitionObject.nullDrop(pandasDf)))
    
    return sparkDataFrame2


# In[ ]:







#!/usr/bin/env python
# coding: utf-8

# In[30]:


from rdflib import Graph, Literal, Namespace, URIRef,BNode
from datetime import datetime


# In[2]:


class Semantification:
    """
    The Semantification class is used to semantify data by creating an RDF graph and serializing it to a destination file.

    Attributes:
    - _graph (rdflib.Graph): The RDF graph to store the semantified data.

    Methods:
    - __init__(): Initializes the Semantification class by creating an empty RDF graph.
    - get_graph(): Retrieves the RDF graph.
    - set_graph(graph): Sets the RDF graph to the specified graph.
    - semantify(df, uri1='uri', label1='label', prediction1='prediction', dest='experiment.ttl'): Semantifies the given DataFrame by adding the data to the RDF graph and serializing it.
    """
       
    def __init__(self):     
        # Create an RDF graph
        self._graph = Graph()
        
    # getter functions
    def get_graph(self):
        return self._graph
    
    # setter functions
    def set_graph(self, graph):
        self._graph= graph
   

    
    def semantify(self, df, uri1='uri', label1='label', prediction1='prediction', dest='experiment.ttl'):
        
        """
        Semantifies the given DataFrame by creating an RDF graph and serializing it to a destination file.

        Parameters:
        - df (pyspark.sql.DataFrame): The DataFrame containing the data to semantify.
        - uri1 (str): The column name in the DataFrame representing the URI of each instance.
        - label1 (str): The column name in the DataFrame representing the label of each instance.
        - prediction1 (str): The column name in the DataFrame representing the prediction of each instance.
        - dest (str): The destination file path where the RDF graph will be serialized. Default is 'experiment.ttl'.

        Returns:
        None
        
        """
        
        # Define namespaces
        EX_NS = Namespace("http://example.com/")
        self._graph.bind("my", EX_NS)
        
        # Define experiment type properties
        experimentTypePropertyURIasString = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
        experimentTypeURIasString = "https://example/Vocab/experiment"
        # Create experiment node
        metagraphDatetime = datetime.now()
        experimentHash = str(hash(str(metagraphDatetime)))
        experimentNode = BNode(experimentHash)
        # Create experiment type properties nodes
        experimentTypePropertyNode = URIRef(experimentTypePropertyURIasString)
        experimentTypeNode = URIRef(experimentTypeURIasString)

        # Add experiment type triple
        self._graph.add((experimentNode, experimentTypePropertyNode, experimentTypeNode))
       
        for row in df.rdd.collect():
            # Get the values from the DataFrame row
            uri = row[uri1]
            label = row[label1]
            prediction = row[prediction1]

            # Create a URIRef for the instance node
            instance_node = URIRef(uri)

            # Add the instance node to the graph
            #graph.add((instance_node, RDF.type, FOAF.Person))

            # Add the label and prediction relations
            self._graph.add((instance_node, EX_NS.label, Literal(label)))
            self._graph.add((instance_node, EX_NS.prediction, Literal(prediction)))
            self._graph.add((experimentNode, EX_NS.hasPrediction, Literal(prediction)))
            
        
        self._graph.serialize(destination=dest)      




#!/usr/bin/env python
# coding: utf-8

# In[1]:


#import findspark
#findspark.find()
#findspark.init()
#from pyspark.sql import SparkSession


# In[2]:


import sparql_dataframe
import pandas as pd


# In[14]:


class DataAcquisition:
    
    """
    A class for knowdleg graph query and data preprocessing tasks such as null handling and null Drop.

    Attributes:
        _endpoint (str): The endpoint for retrieving data.
        _query (str): The query for retrieving data.
        _amputationMethod (str): The method for handling null values in the DataFrame ('nullReplacement' or 'nullDrop').
        _rowNullDropPercent (int): The percentage threshold for dropping rows with null values.
        _columnNullDropPercent (int): The percentage threshold for dropping columns with null values.
        _nullReplacementMethod (str): The method for replacing null values ('median', 'mean', 'mode', or 'customValue').
        _customValueVariable (str): The name of the variable used as a custom replacement value for null values.
        _customStringValueVariable (str): The name of the variable used as a custom replacement string for null values.

    """

    sparkSession = None
    #spark = SparkSession.builder.getOrCreate()
     
    def __init__(self, sparkSession):

        DataAcquisition.sparkSession = sparkSession
   
        self._endpoint = ''
        self._query = ''
        
       
        self._amputationMethod='nullDrop' # nullReplacement|nullDropp
        self._rowNullDropPercent=100 # it is percent %
        self._columnNullDropPercent=0 # it is percent %
        self._nullReplacementMethod='mod' # median|mean|mode|customValue
        self._customValueVariable=''
        self._customStringValueVariable=''
        
    
    # getter functions
    
    def get_endpoint(self):
        return self._endpoint
    
    def get_query(self):
        return self._query
    
    def get_amputationMethod(self):
        return self._amputationMethod
    
    def get_rowNullDropPercent(self):
        return self._rowNullDropPercent
    
    def get_columnNullDropPercent(self):
        return self._columnNullDropPercent
    
    def get_nullReplacementMethod(self):
        return self._nullReplacementMethod
    
    def get_customValueVariable(self):
        return self._customValueVariable
    
    def get_customStringValueVariable(self):
        return self._customStringValueVariable
    

    # setter functions
    
    def set_endpoint(self, endpoint):
        self._endpoint= endpoint
        
    def set_query(self, query):
        self._query= query 
    
    def set_amputationMethod(self, amputationMethod):
        self._amputationMethod=amputationMethod
    
    def set_rowNullDropPercent(self, rowNullDropPercent):
        self._rowNullDropPercent=rowNullDropPercent
    
    def set_columnNullDropPercent(self, columnNullDropPercent):
        self._columnNullDropPercent=columnNullDropPercent
    
    def set_nullReplacementMethod(self, nullReplacementMethod):
        self._nullReplacementMethod= nullReplacementMethod
        
    def set_customValueVariable(self, customValueVariable):
        self._customValueVariable= customValueVariable
        
    def set_customStringValueVariable(self, customStringValueVariable):
        self._customStringValueVariable= customStringValueVariable
        
        
    def getDataFrame(self, endpoint=None, query=None):
        """Retrieve data from a SPARQL endpoint and convert it into a Spark DataFrame.

        Args:
            endpoint (str, optional): The SPARQL endpoint URL. If not provided, the default endpoint will be used.
            query (str, optional): The SPARQL query string. If not provided, the default query will be used.

        Returns:
            pyspark.sql.DataFrame: The resulting Spark DataFrame.

        Raises:
            TypeError: If there are null values in the Pandas DataFrame and no handling method is specified.

        Notes:
            This function retrieves data from a SPARQL endpoint and converts it into a Spark DataFrame. It follows the
            following steps:

            1. If the endpoint is not provided, the default endpoint is used. If the default endpoint is not set, an error
               message is displayed and the function returns.

            2. If the query is not provided, the default query is used. If the default query is not set, an error message
               is displayed and the function returns.

            3. The data is queried from the SPARQL endpoint and converted into a Pandas DataFrame.

            4. If there are null values in the Pandas DataFrame, handling methods are applied based on the configured
               amputation method.

            5. The Pandas DataFrame is then converted into a Spark DataFrame.

            6. The resulting Spark DataFrame is returned.
        """
        
        
        #(Sparql-->PandasDF-->SparkDf)
        
        if endpoint==None:
            endpoint=self._endpoint
            if endpoint == '':
                print('Can not proceed further, please provide the enpoint')
                return
            
        if query==None:
            query=self._query
            if query == '':
                print('Can not proceed further, please provide the query')
                return
        
        #query from sparql endpoint and then create Pandas dataframe 
        pandasDataFrame=(sparql_dataframe.get(endpoint, query))
        #create Spark Dataframe from Pandas Dataframe, if there is Null values, go to handling methods
        try:
            sparkDataFrame= DataAcquisition.sparkSession.createDataFrame(pandasDataFrame)
        except TypeError as error:
            print(error)
            print('Null values exist, handling methods will be applied')
                    
            if self._amputationMethod=='nullDrop':
                sparkDataFrame= DataAcquisition.sparkSession.createDataFrame(self.nullDrop(pandasDataFrame))
            
            if self._amputationMethod=='nullReplacement':
                sparkDataFrame= DataAcquisition.sparkSession.createDataFrame(self.nullReplacement(pandasDataFrame))
                
            
        return sparkDataFrame
    
    
    def nullReplacement(self, df):
        """Apply null replacement methods on variables with NaN values in a DataFrame.
    
        Args:
            df (pandas.DataFrame): The input DataFrame.

        Returns:
            pandas.DataFrame: The DataFrame with null values replaced according to the specified method.

        Raises:
            ValueError: If the customStringValueVariable or customValueVariable is not defined for the 'customValue'
                        null replacement method.

        Notes:
            This function applies null replacement methods on variables with NaN values in the input DataFrame. It follows
            the following steps:

            1. If the null replacement method is set to 'median', iterate over the columns of the DataFrame that have
               null values and fill them with the column's median value. Note that this method cannot be applied to string
               columns.

            2. If the null replacement method is set to 'mean', iterate over the columns of the DataFrame that have
               null values and fill them with the column's mean value.

            3. If the null replacement method is set to 'mod' (mode/most frequent value), iterate over the columns of the
               DataFrame that have null values and fill them with the column's mode (first most frequent value). Note that
               this method cannot be applied to string columns.

            4. If the null replacement method is set to 'customValue', iterate over the columns of the DataFrame that have
               null values. If the column's data type is an object (string), fill the null values with the specified custom
               string value. If the column's data type is not an object, fill the null values with the specified custom
               numeric value.

            5. The resulting DataFrame with replaced null values is returned.
        """
        
        #fill with median
        if self._nullReplacementMethod=="median":
            for i in df.columns[df.isnull().any(axis=0)]:     
                df[i].fillna(df[i].median(),inplace=True)
        
        #fill with mean
        if self._nullReplacementMethod=="mean":
            for i in df.columns[df.isnull().any(axis=0)]:     
                df[i].fillna(df[i].mean(),inplace=True)
        
        #fill with mode(most frequent value)
        if self._nullReplacementMethod=="mod":
            for i in df.columns[df.isnull().any(axis=0)]:     
                df[i].fillna(df[i].mode().iloc[0],inplace=True)
                
        #fill with user defined value
        if self._nullReplacementMethod=="customValue":
            for i in df.columns[df.isnull().any(axis=0)]:
                if df[i].dtypes == object:
                    if self._customStringValueVariable=='':
                        print('Please define the customStringValueVariable or change the nullReplacementMethod')
                        return
                    df[i].fillna(self._customStringValueVariable,inplace=True)
                    print(f'NaN values in -{i}- column replaced by value -{self._customStringValueVariable}-')
                else:
                    if self._customValueVariable=='':
                        print('Please define the customValueVariable or change the nullReplacementMethod')
                        return
                    df[i].fillna(self._customValueVariable,inplace=True)
                    print(f'NaN values in -{i}- column replaced by value -{self._customValueVariable}-')
  
        return df
    
    
    def nullDrop(self, df):
        """Apply null dropping according to thresholds on a DataFrame.
    
        Args:
            df (pandas.DataFrame): The input DataFrame.

        Returns:
            pandas.DataFrame: The DataFrame with null values dropped according to the specified thresholds.

        Notes:
            This function applies null dropping on the input DataFrame based on the following steps:

            1. Drop columns where the percentage of missing values is greater than or equal to the specified
               `self._columnNullDropPercent` threshold.

            2. Drop rows where the percentage of missing values is greater than or equal to the specified
               `self._rowNullDropPercent` threshold.

            3. If there are still null values in the DataFrame after dropping, the nullReplacement function is called
               to apply the specified null replacement method.

            4. The resulting DataFrame with dropped null values is returned.

        Warnings:
            - The nullReplacement method will be called if there are still null values after dropping. Make sure the
              nullReplacementMethod is properly configured.

        """
        
        #first: column base, if null values are over a threshold percent
        print(f'Drop the columns where at least %{self._columnNullDropPercent} element is missing.')
        df.dropna(axis='columns',thresh=(len(df)*(self._columnNullDropPercent/100)),inplace = True)
        
        #second: row base, if null values are over a threshold percent
        print(f'Drop the rows where at least %{self._rowNullDropPercent} element is missing.')
        df.dropna(axis='index',thresh=(len(df.columns)*(self._rowNullDropPercent/100)),inplace = True)
        
        #if there are still null values, call the nullReplacement function
        if [i for i in df.columns[df.isnull().any(axis=0)]]:
            print(f'After dropping, there are still null values. {self._nullReplacementMethod} null replacement method will be applied')
            df=self.nullReplacement(df)
        
        return df  




#!/usr/bin/env python
# coding: utf-8

# In[31]:


#import findspark
#findspark.find()
#findspark.init()


# In[32]:


#from pyspark.sql import SparkSession
from pyspark.ml.feature import StringIndexer,Tokenizer,StopWordsRemover,Word2Vec,VectorAssembler
from pyspark.sql.types import StringType,IntegerType,FloatType,DoubleType,LongType,BooleanType
#from pyspark.sql.functions import collect_list,size,col,explode_outer
from pyspark.sql.functions import *


# In[33]:


class Vectorization:
    
    """
    A class for data preprocessing tasks such as null handling, feature extraction, and vectorization.

    Attributes:
        _entityColumn (str): The name of the entity column.
        _labelColumn (str): The name of the label column.
        _features (list): A list of features containing information about each column.
        _stopWordsRemover (bool): A flag indicating whether to remove stop words during word embedding.
        _word2vecSize (int): The size of the word vectors in Word2Vec embedding.
        _word2vecMinCount (int): The minimum count of words required for Word2Vec embedding.
        _digitStringStrategy (str): The strategy for digitizing string values ('index' or 'hash').

    """

    #sparkSession = None 
    #spark = SparkSession.builder.getOrCreate()
     
    def __init__(self):

        #Vectorization.sparkSession=sparkSession
        self._entityColumn = ''
        self._stopWordsRemover=True        
        self._word2vecSize=2
        self._word2vecMinCount=1
        self._digitStringStrategy = "index" # index | hash
        
    
    # getter functions
    
    def get_entityColumn(self):
        return self._entityColumn
    
    def get_word2vecSize(self):
        return self._word2vecSize
    
    def get_stopWordsRemover(self):
        return self._StopWordsRemover
    
    def get_word2vecMinCount(self):
        return self._word2vecMinCount
    
    def get_digitStringStrategy(self):
        return self._digitStringStrategy
    
    
    # setter functions
    
    def set_entityColumn(self, entityColumn):
        self._entityColumn= entityColumn
        
    def set_word2vecSize(self, word2vecSize):
        self._word2vecSize= word2vecSize
        
    def set_word2vecMinCount(self, word2vecMinCount):
        self._word2vecMinCount= word2vecMinCount
        
    def set_stopWordsRemover(self, StopWordsRemover):
        self._StopWordsRemover= StopWordsRemover
        
    def set_digitStringStrategy(self, digitStringStrategy):
        self._digitStringStrategy= digitStringStrategy
        
    
    
    def vectorize(self, df2, features):
        """
        Vectorizes the specified columns in the DataFrame based on the provided features.

        Args:
            df2 (pyspark.sql.DataFrame): The input DataFrame.
            features (dict): A dictionary containing information about the features of each column.

        Returns:
            fullDigitizedDf (pyspark.sql.DataFrame): The vectorized DataFrame.

        Raises:
            NotImplementedError: If a transformation for a specific data type is not implemented.

        Implementation Flow:
            1. Iterate over each column in the DataFrame.

            2. Check the data type and features of the column to determine the vectorization strategy.

            3. Apply the appropriate transformation based on the column's data type and features.

               - If the column is a Single Categorical String:
                   - Apply string indexing or hashing based on the configured strategy.

               - If the column is a List of Categorical Strings:
                   - Explode the list and apply string indexing or hashing based on the configured strategy.

               - If the column is a Single Non-Categorical String:
                   - Apply Word2Vec embedding after tokenization and optional stop word removal.

               - If the column is a List of Non-Categorical Strings:
                   - Combine the list elements, apply tokenization, optional stop word removal, and Word2Vec embedding.

               - If the column is a Numeric type (Integer, Long, Float, Double):
                   - Handle both Single and List types by either joining or exploding the values.

               - If the column is a Boolean type:
                   - Cast the Boolean values to Integer (0 or 1).

               - If the column is of an unsupported data type:
                   - Raise a NotImplementedError.

            4. Join the transformed column with the vectorized DataFrame using the entity column.

            5. Return the resulting vectorized DataFrame.

        Note:
            The implementation follows a conditional branching based on the data type and features of each column
            to determine the appropriate vectorization strategy.

        """
        #if entity column is not set, set it as the first column
        if self._entityColumn == '':
            self._entityColumn=df2.columns[0]
            print(f'No entity column has been set, that is why the first column {self._entityColumn} is used as entity column')
            
        fullDigitizedDf=df2.select(self._entityColumn)
        
        for columnName in [col for col in df2.columns if col != self._entityColumn]:
    
            if features[columnName]['datatype']==StringType():

                #Single_Categorical_String
                if features[columnName]['isListOfEntries']==False and features[columnName]['isCategorical']==True:
                    
                    if self._digitStringStrategy == "index":

                        #newFeatureColumnName = columnName + "(IndexedString)"

                        indexer = StringIndexer().setInputCol(columnName).setOutputCol("output")
                        digitizedDf = indexer.fit(df2).transform(df2).select(self._entityColumn, "output")

                        fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn).withColumnRenamed("output", columnName)
                        
                    elif self._digitStringStrategy == "hash":
                        
                        hashedDf = df2.withColumn("output", hash(col(columnName)).cast("double")) \
                                       .select(self._entityColumn, "output")
        

                        fullDigitizedDf = fullDigitizedDf.join(hashedDf,self._entityColumn).withColumnRenamed("output", columnName)

                #ListOf_Categorical_String
                elif features[columnName]['isListOfEntries']==True and features[columnName]['isCategorical']==True:

                    #newFeatureColumnName = columnName + "(ListOfIndexedString)"

                    #transform list to single  by exploding
                    inputDf = df2.select(col(self._entityColumn), explode_outer(col(columnName)))
                    
                    if self._digitStringStrategy == "index":

                        indexer = StringIndexer().setInputCol("col").setOutputCol("outputTmp")

                        digitizedDf = indexer.fit(inputDf).transform(inputDf) \
                                .select(self._entityColumn, "outputTmp") \

                        '''
                        digitizedDf = indexer.fit(inputDf).transform(inputDf) \
                                .groupBy(_entityColumn) \
                                .agg(collect_list("outputTmp").alias("output")) \
                                .select(_entityColumn, "output") \
                                .withColumnRenamed("output", newFeatureColumnName) \
                                .select(_entityColumn, newFeatureColumnName)
                        '''

                        fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn).withColumnRenamed("outputTmp", columnName)
                
                    elif self._digitStringStrategy == "hash":
                        
                        hashedDf = inputDf.withColumn("output", hash(col("col")).cast("double")) \
                                       .select(self._entityColumn, "output")
                
                        fullDigitizedDf = fullDigitizedDf.join(hashedDf,self._entityColumn).withColumnRenamed("output", columnName)
                
                #Single_NonCategorical_String
                elif features[columnName]['isListOfEntries']==False and features[columnName]['isCategorical']==False:

                    #newFeatureColumnName = columnName +"(Word2Vec)"
                    
                    tokenizer = Tokenizer(inputCol=columnName, outputCol="words")
                    df2 = tokenizer.transform(df2)
                    tempColumnName="words"
                    
                    if self._stopWordsRemover==True:

                        remover = StopWordsRemover(inputCol="words", outputCol="filtered")
                        df2 = remover.transform(df2)
                        tempColumnName='filtered'

                    word2vec = Word2Vec(inputCol=tempColumnName, outputCol="output", minCount=self._word2vecMinCount, vectorSize=self._word2vecSize)
                    word2vecModel = word2vec.fit(df2)
                    digitizedDf=word2vecModel.transform(df2) \
                                             .select(self._entityColumn, "output")

                    fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn).withColumnRenamed("output", columnName)


                #ListOf_NonCategorical_String
                elif features[columnName]['isListOfEntries']==True and features[columnName]['isCategorical']==False:

                    #newFeatureColumnName = columnName +"(Word2Vec)"

                    dfCollapsedTwoColumnsNullsReplaced = df2 \
                        .withColumn("sentences", concat_ws(". ", col(columnName))) \
                        .select(self._entityColumn, "sentences")
                        
                    tokenizer = Tokenizer().setInputCol("sentences").setOutputCol("words")
                    tokenizedDf = tokenizer.transform(dfCollapsedTwoColumnsNullsReplaced).select(self._entityColumn, "words")
                    tempColumnName='words'    
                        
                    if self._stopWordsRemover==True:

                        remover = StopWordsRemover().setInputCol("words").setOutputCol("filtered")
                        tokenizedDf = remover.transform(tokenizedDf).select(self._entityColumn, "filtered")
                        tempColumnName='filtered'
                        
                    word2vec = Word2Vec(inputCol=tempColumnName, outputCol="output", minCount=self._word2vecMinCount, vectorSize=self._word2vecSize)    
                    word2vecModel = word2vec.fit(tokenizedDf)
                    digitizedDf=word2vecModel.transform(tokenizedDf)\
                                            .select(self._entityColumn, "output")
                        
                    fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn).withColumnRenamed("output", columnName)


            elif features[columnName]['datatype']==IntegerType() or features[columnName]['datatype']==LongType() or \
                 features[columnName]['datatype']==FloatType() or features[columnName]['datatype']==DoubleType():

                 #if list, need to convert to single value
                if features[columnName]['isListOfEntries']==True:
                        digitizedDf=df2.select(self._entityColumn,explode_outer(columnName).alias(columnName))
                        fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn)  

                #if single, add to df
                else:
                    digitizedDf=df2.select(self._entityColumn,columnName)
                    fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn)

            elif features[columnName]['datatype']==BooleanType():
                
                digitizedDf=df2.select(self._entityColumn,columnName)
                #casting Boolean Type to Integer where it converts values to 0's and 1's
                digitizedDf=digitizedDf.withColumn(columnName,col(columnName).cast(IntegerType()))
                
                fullDigitizedDf = fullDigitizedDf.join(digitizedDf,self._entityColumn)
                
            #if it is not String     
            else:
                type1=features[columnName]['datatype']
                print( f'Column {columnName} can not be added. Transformation for type {type1} not implemented yet.')
              
        
        
        return fullDigitizedDf

